package jj2000.j2k.encoder;

import jj2000.j2k.entropy.encoder.WPRateAllocator;
import jj2000.j2k.entropy.encoder.RDResult;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;
import jj2000.j2k.util.ParameterList;
import jj2000.j2k.entropy.encoder.SubbandRateDistStats;


public class WPEncoderBuilder {
	
	public static Encoder createEncoder(ParameterList pl) {
		Encoder enc;
		if (pl.getBooleanParameter("Bwps")) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, 
					"Create RDWPEncoder.");
			enc = WPEncoderBuilder.createRDOWPEncoder(pl);
		} else {
			enc = new Encoder(pl);
		}
		return enc;
	}

	/**
	 * Returns an encoder that employs a rate-distortion-optimal wavelet packet
	 * basis (RDO-WPB). The function determines the RDO-WPB, i.e., searches for
	 * the best basis for a given rate and returns an Encoder with this
	 * basis. 
	 */
	public static Encoder createRDOWPEncoder(ParameterList pl) {
		int tile = 0;
		int component = 0;
		
		// Check whether header data is taken into account, i.e., set bHeader.
		
		// Bottom up optimization or top down.
		boolean bBottomUp = pl.getBooleanParameter("Bbottom_up");
		// Do we equalize the distortion of parent and children?
		boolean bDistEqualize = pl.getBooleanParameter("Bequalize");
		// Default: All wavelet packet bases are considered and not just
		// the allowed bases of JPEG2000 Part 2.
		boolean bJ2KP2 = pl.getBooleanParameter("Bj2kp2");
		// Use suband rate distortion statistics.
		boolean bDirect = pl.getBooleanParameter("Bdirect");
		// Add a maximum wavelet decomposition depth (required).
		int maxWlev = pl.getIntParameter("Wlev");

		if(pl.getBooleanParameter("debug"))
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,  
			"Setting bBottomUp to " + bBottomUp + "\n" 
			+ "Setting bDistEqualize to " + bDistEqualize + "\n" 
			+ "Setting bJ2KP2 to " + bJ2KP2 + "\n" 
			+ "Performing selecting RDO WP for maximum wavelet decompostion " + maxWlev
			);
		}
		/*
		 * Find the best WPB in RD sense.
		 */
		// Get a standard Encoder.
		Encoder enc = new Encoder(pl);

		// Get full decompositions for all wavelet decomposition levels.
		WPRateAllocator[] fullDecompRateAllocators = new WPRateAllocator[maxWlev + 1];
		for (int w = 0; w <= maxWlev; w++) {
			fullDecompRateAllocators[w] = enc.getWPRateAllocator(w);
			if (fullDecompRateAllocators[w] == null) {
				System.out
						.println("Can not instantiate PostCompRateAllocator. Check input parameters.");
				System.exit(-1);
			}
		}
 
		if (bDistEqualize) {
			int bHeader = pl.getIntParameter("Aheader_alg");
			equalizeDists(fullDecompRateAllocators, 0, 0, bHeader);
		}

		
		int availableBytes = fullDecompRateAllocators[0].getAvailableBytes();
		int mainAndTilePartHeaderSize = fullDecompRateAllocators[0].getMainHeaderLength();
		mainAndTilePartHeaderSize += fullDecompRateAllocators[0].getTilePartHeaderLength(
				tile, component);
		System.out.println("Final: mainAndTilePartHeaderSize "
				+ mainAndTilePartHeaderSize);
		System.out.println("Available bytes = " + availableBytes);

		RDResult rdj;
		if (!pl.getBooleanParameter("lossless")) {
			// Lossy.
			rdj = findWPBforRate(fullDecompRateAllocators, availableBytes,
					mainAndTilePartHeaderSize, bJ2KP2, bBottomUp);
		} else {
			// Lossless (only best basis / top down implemented).
			rdj = findBestWpbforLossless(fullDecompRateAllocators, 0, 0, bJ2KP2);
		}

		/*
		 * Encode image with the best WPB, i.e., create a standard encoder with
		 * the best WPB.
		 */
		pl.setProperty("Wwps", "b" + rdj.wpsDesc);
		System.out.println("Encode with the following WPB = " + rdj.wpsDesc);
		pl.setProperty("Wlev", Integer.toString(getMaxLLDecomp(rdj.wpsDesc)));
		if (bDirect) {
			System.out.println("Bdirect on -> Athresh");
			pl.setProperty("Athresh", Double.toString(rdj.lambda));
		} else {
			System.out.println("Bdirect off -> Asb_thresh.");
			pl.setProperty("Asb_thresh", Double.toString(rdj.lambda));
		}
		enc = new Encoder(pl);
		return enc;

	}

	// Equalize distortion between children and parents with max distortion
	// (BottomUp).
	private static void equalizeDists(WPRateAllocator[] ras, int wLev,
			int sbIdx, int headerStrategy) {
		if (wLev + 1 >= ras.length)
			return;

		equalizeDists(ras, wLev + 1, 4 * sbIdx, headerStrategy);
		equalizeDists(ras, wLev + 1, 4 * sbIdx + 1, headerStrategy);
		equalizeDists(ras, wLev + 1, 4 * sbIdx + 2, headerStrategy);
		equalizeDists(ras, wLev + 1, 4 * sbIdx + 3, headerStrategy);

		int resOfSbb = WPRateAllocator.getResLevelFromSubbandIndex(sbIdx);
		SubbandRateDistStats parentSRD, child0SRD, child1SRD, child2SRD, child3SRD;
		parentSRD = ras[wLev].getRDStatOfSubband(0, 0, sbIdx, resOfSbb,
				headerStrategy);

		resOfSbb = WPRateAllocator.getResLevelFromSubbandIndex(4 * sbIdx);
		child0SRD = ras[wLev + 1].getRDStatOfSubband(0, 0, 4 * sbIdx, resOfSbb,
				headerStrategy);

		resOfSbb = WPRateAllocator.getResLevelFromSubbandIndex(4 * sbIdx + 1);
		child1SRD = ras[wLev + 1].getRDStatOfSubband(0, 0, 4 * sbIdx + 1,
				resOfSbb, headerStrategy);

		resOfSbb = WPRateAllocator.getResLevelFromSubbandIndex(4 * sbIdx + 2);
		child2SRD = ras[wLev + 1].getRDStatOfSubband(0, 0, 4 * sbIdx + 2,
				resOfSbb, headerStrategy);

		resOfSbb = WPRateAllocator.getResLevelFromSubbandIndex(4 * sbIdx + 3);
		child3SRD = ras[wLev + 1].getRDStatOfSubband(0, 0, 4 * sbIdx + 3,
				resOfSbb, headerStrategy);

		double parentMaxDist = parentSRD.dists.get(0);
		double childrenMaxDist = child0SRD.dists.get(0)
				+ child1SRD.dists.get(0) + child2SRD.dists.get(0)
				+ child3SRD.dists.get(0);
		double ptc = parentMaxDist / childrenMaxDist;

		parentSRD.multDists(1. / ptc);
	}
	
	/**
	 * 
	 * Search for the best matching lambda for a certain file size (in bytes).
	 * The best WPB is chosen in dependency of lambda, i.e, the file size.
	 * 
	 * @param fullDecompWPRateAlloctors
	 * @param maxBytes
	 * @param headerOverhead
	 * @param bHeader
	 * @param bJ2KP2
	 * @param bBottomUp
	 * @return
	 */
	private static RDResult findWPBforRate(WPRateAllocator[] fullDecompWPRateAlloctors,
			int maxBytes, int headerOverhead, boolean bJ2KP2, boolean bBottomUp) {
		
		// The maximum number of searches in the bisection algorithm.
		int availableBytes = maxBytes - headerOverhead;
		int maxSearchSteps = 32;

		float LAMBDA_EPSILON = (float) Math.pow(10., -9.);
		float slambda = 0; // Float.MIN_VALUE;
		float hlambda = 10;
		float lambda = 5f;

		RDResult rdres;
		RDResult finalrdres = new RDResult();

		/**
		 * DEBUG if(true){ rdres = findBestWPBforLambda(allocarray, 0, 0, 0.11f,
		 * bHeader, bJ2KP2); System.out.println("The packet data is " +
		 * rdres.rate); return rdres; } END DEBUG
		 **/

		int numSearchSteps = 0;
		while (numSearchSteps < maxSearchSteps) {
			numSearchSteps++;

			if (bBottomUp) {
				rdres = findBestWpbBottomUp(fullDecompWPRateAlloctors, 0, 0, lambda, bJ2KP2);
			} else {
				rdres = findGoodWpbTopDown(fullDecompWPRateAlloctors, 0, 0, lambda, bJ2KP2);
			}

			if (rdres.rate == availableBytes) {
				finalrdres = rdres;
				break;
			}
			if (rdres.rate < availableBytes) {
				// Save the best result, which has a smaller rate than the
				// target rate.
				if (finalrdres.rate < rdres.rate) {
					finalrdres = rdres;
				}
				if (Math.abs(hlambda - lambda) < LAMBDA_EPSILON)
					break;
				hlambda = lambda;
				lambda = slambda + (hlambda - slambda) / 2;
			} else if (rdres.rate > availableBytes) {
				if (Math.abs(slambda - lambda) < LAMBDA_EPSILON)
					break;
				slambda = lambda;
				lambda = slambda + (hlambda - slambda) / 2;
			}
			System.out.println("Lambda = " + rdres.lambda + " with a rate of "
					+ rdres.rate + " and an estimated file size of "
					+ (finalrdres.rate + headerOverhead + 2)
					+ " for a maximum of " + maxBytes
					+ " with steps of iteration = " + numSearchSteps);

		}
		int EOC = 2;
		System.out.println("Final lambda = " + finalrdres.lambda
				+ " with a final rate of " + finalrdres.rate
				+ " and an estimated file size of "
				+ (finalrdres.rate + headerOverhead + EOC)
				+ " for a maximum of " + maxBytes
				+ " with steps of iteration = " + numSearchSteps);
		return finalrdres;
	}

	/**
	 * findBestWPBforLambda( rateAllocators[], 0, 0, lambda, bHeader) returns
	 * the best WPB for a given lambda. The method is recursively called to
	 * determine the cost of a subband until wLev equals the maximum
	 * decomposition depth.
	 * 
	 * @param rateAllocators
	 *            The array of rate allocators containing all RD-information of
	 *            full wavelet packet decompositions for depth 0 to
	 *            aralloc.length-1.
	 * @param wLev
	 *            The considered decomposition depth.
	 * @param sbIdx
	 *            The subband which is analyzed (split or merged).
	 * @param lambda
	 *            The targeted lambda, i.e. rdThreshold.
	 * @param bHeader
	 *            Shall we take the packet header data into account?
	 * @param bJ2KP2
	 *            Shall we restrict the bases to the ones of JPEG2000 Part 2?
	 * @return The best wavelet packet basis and its associated costs and
	 *         lambda.
	 */
	private static RDResult findBestWpbBottomUp(WPRateAllocator[] rateAllocators,
			int wLev, int sbIdx, float lambda, boolean bJ2KP2) {
		// System.out.println(wLev + ", " + sbIdx +
		// ", Finding best basis for rdTreshold = "+ lambda);

		RDResult parentRDJ;

		if (bJ2KP2 && (sbIdx >= 64)) { // J2K Part 2 Wavelet Bases
			double rate = Double.MAX_VALUE;
			double dist = Double.MAX_VALUE;

			parentRDJ = new RDResult(rate, dist, lambda);

		} else { // All bases and header data
			parentRDJ = (rateAllocators[wLev]).getRDCostForSubband(sbIdx,
					lambda);
		}

		if (wLev == rateAllocators.length - 1) {
			return new RDResult(parentRDJ, "0");
		}

		// The children of a subband sb (with index sbIdx) are indexed from
		// 4*sbIdx to 4*sbIdx+3.
		// This index scheme allows to distinguish the subbands of a resolution.
		// If a full wavelet packet decomposition is done, the index scheme
		// leads to unique indices for all subbands.
		int childSbIdx = sbIdx * 4;
		RDResult child0 = findBestWpbBottomUp(rateAllocators, wLev + 1,
				childSbIdx + 0, lambda, bJ2KP2);
		RDResult child1 = findBestWpbBottomUp(rateAllocators, wLev + 1,
				childSbIdx + 1, lambda, bJ2KP2);
		RDResult child2 = findBestWpbBottomUp(rateAllocators, wLev + 1,
				childSbIdx + 2, lambda, bJ2KP2);
		RDResult child3 = findBestWpbBottomUp(rateAllocators, wLev + 1,
				childSbIdx + 3, lambda, bJ2KP2);

		double childrenCost = child0.cost + child1.cost + child2.cost
				+ child3.cost;

		/**
		 * DEBUG System.out.println("Finding best basis for rdTreshold = "+
		 * lambda); System.out.println("Subband (wlev, sbIdx) = ("+ wLev + ", "
		 * + sbIdx + ") has a cost of " + parentRDJ.cost +
		 * " and a children cost of " + childrenCost + "."); END DEBUG
		 */
		if (parentRDJ.cost > childrenCost) {
			// Split decision
			double childrenDist = child0.dist + child1.dist + child2.dist
					+ child3.dist;
			double childrenRate = child0.rate + child1.rate + child2.rate
					+ child3.rate;

			return new RDResult(childrenRate, childrenDist, childrenCost,
					lambda, "1" + child0.wpsDesc + child1.wpsDesc
							+ child2.wpsDesc + child3.wpsDesc);
		} else {
			// Merge decision
			return new RDResult(parentRDJ, "0");
		}
	}

	private static RDResult getSubbandCost(WPRateAllocator[] rateAllocators,
			int wLev, int sbIdx, float lambda, boolean bJ2KP2) {
		RDResult parentRDJ;

		if (bJ2KP2 && (sbIdx >= 64)) { // J2K Part 2 Wavelet Bases
			double rate = Double.MAX_VALUE;
			double dist = Double.MAX_VALUE;

			parentRDJ = new RDResult(rate, dist, lambda);

		} else { // All bases
			parentRDJ = (rateAllocators[wLev]).getRDCostForSubband(sbIdx,
					lambda);
		}
		return parentRDJ;
	}

	/**
	 * Start at the root. Decompose a subband only if its children cost is
	 * smaller than its cost. If decompose then check each child.
	 */
	private static RDResult findGoodWpbTopDown(
			WPRateAllocator[] rateAllocators, int wLev, int sbIdx,
			float lambda, boolean bJ2KP2) {
		// Determine cost of current subband.
		RDResult parentRDJ = getSubbandCost(rateAllocators, wLev, sbIdx,
				lambda, bJ2KP2);
		// At the highest decomposition level we have no more children.
		if (wLev == rateAllocators.length - 1) {
			return new RDResult(parentRDJ, "0");
		}

		// Determine cost of children
		int childSbIdx = sbIdx * 4;
		RDResult child0 = getSubbandCost(rateAllocators, wLev + 1,
				childSbIdx + 0, lambda, bJ2KP2);
		RDResult child1 = getSubbandCost(rateAllocators, wLev + 1,
				childSbIdx + 1, lambda, bJ2KP2);
		RDResult child2 = getSubbandCost(rateAllocators, wLev + 1,
				childSbIdx + 2, lambda, bJ2KP2);
		RDResult child3 = getSubbandCost(rateAllocators, wLev + 1,
				childSbIdx + 3, lambda, bJ2KP2);

		double childrenCost = child0.cost + child1.cost + child2.cost
				+ child3.cost;

		if (parentRDJ.cost > childrenCost) {
			// Split decision
			// Decompose further
			child0 = findGoodWpbTopDown(rateAllocators, wLev + 1,
					childSbIdx + 0, lambda, bJ2KP2);
			child1 = findGoodWpbTopDown(rateAllocators, wLev + 1,
					childSbIdx + 1, lambda, bJ2KP2);
			child2 = findGoodWpbTopDown(rateAllocators, wLev + 1,
					childSbIdx + 2, lambda, bJ2KP2);
			child3 = findGoodWpbTopDown(rateAllocators, wLev + 1,
					childSbIdx + 3, lambda, bJ2KP2);

			childrenCost = child0.cost + child1.cost + child2.cost
					+ child3.cost;

			double childrenDist = child0.dist + child1.dist + child2.dist
					+ child3.dist;
			double childrenRate = child0.rate + child1.rate + child2.rate
					+ child3.rate;

			return new RDResult(childrenRate, childrenDist, childrenCost,
					lambda, "1" + child0.wpsDesc + child1.wpsDesc
							+ child2.wpsDesc + child3.wpsDesc);
		} else {
			// Merge decision
			return new RDResult(parentRDJ, "0");
		}
	}

	// We only consider the rate and assume that the transform is lossless.
	// This direct approach prevents lambda iteration and instabilities around
	// zero for the rdThreshold lambda (in the lossy algorithm).
	private static RDResult findBestWpbforLossless(
			WPRateAllocator[] rateAllocators, int wLev, int sbIdx, boolean bJ2KP2) {
	
		float lambda = 0;
		RDResult parentRDJ;
	
		if (bJ2KP2 && (sbIdx >= 64)) { // J2K Part 2 Wavelet Bases
			double rate = Double.MAX_VALUE;
			double dist = Double.MAX_VALUE;
	
			parentRDJ = new RDResult(rate, dist, lambda);
		} else { // All bases and header data
			parentRDJ = (rateAllocators[wLev]).getRDCostForSubband(sbIdx,
					lambda);
			// .calcRDCostForSubbandForLambdaWithHeader(sbIdx, lambda);
		}
	
		if (wLev == rateAllocators.length - 1) {
			return new RDResult(parentRDJ, "0");
		}
	
		// The children of a subband sb (with index sbIdx) are indexed from
		// 4*sbIdx to 4*sbIdx+3.
		// This index scheme allows to distinguish the subbands of a resolution.
		// If a full wavelet packet decomposition is done, the index scheme
		// leads to unique indices for all subbands.
		int childSbIdx = sbIdx * 4;
		RDResult child0 = findBestWpbforLossless(rateAllocators, wLev + 1,
				childSbIdx + 0, bJ2KP2);
		RDResult child1 = findBestWpbforLossless(rateAllocators, wLev + 1,
				childSbIdx + 1, bJ2KP2);
		RDResult child2 = findBestWpbforLossless(rateAllocators, wLev + 1,
				childSbIdx + 2, bJ2KP2);
		RDResult child3 = findBestWpbforLossless(rateAllocators, wLev + 1,
				childSbIdx + 3, bJ2KP2);
	
		double childrenRate = child0.rate + child1.rate + child2.rate
				+ child3.rate;
	
		/**
		 * DEBUG * System.out.println("Subband (wlev, sbIdx) = ("+ wLev + ", " +
		 * sbIdx + ") has a rate of " + parentRDJ.rate +
		 * " and a children cost of " + childrenCost + "."); END DEBUG
		 */
		if (parentRDJ.rate > childrenRate) {
			// Split decision
			double childrenDist = child0.dist + child1.dist + child2.dist
					+ child3.dist;
			return new RDResult(childrenRate, childrenDist, childrenRate,
					lambda, "1" + child0.wpsDesc + child1.wpsDesc
							+ child2.wpsDesc + child3.wpsDesc);
		} else {
			// Merge decision
			return new RDResult(parentRDJ, "0");
		}
	}

	/**
	 * Returns the maximum decomposition depth of the LL subband of WPS coded as
	 * 1-0-String.
	 * 
	 * @param WPS
	 *            coded as 1-0-String
	 * @return decomposition depth of LL subband
	 * */
	public static int getMaxLLDecomp(String strWps) {
		int ic;
		for (ic = 0; (ic < strWps.length()) && (strWps.charAt(ic) == '1'); ic++) {
			;
		}
		return ic;
	}

	public static void main(String argv[]) {
		ParameterList pl = new ParameterList();
		pl.parseArgs(argv);

		Encoder enc = createRDOWPEncoder(pl);
		enc.run();
	}
}
