package jj2000.j2k.entropy.encoder;

import java.util.Collections;
import java.util.Vector;

/**
 * Summarize the entire rate-distortion information of a subband.
 **/
public class SubbandRateDistStats {
	// n rates.
	public Vector<Double> rates; 
	// n rates.
	public Vector<Double> phbytes; 
	// n dists, the actual distortion (not reduction)!
	public Vector<Double> dists; 
	// n slopes
	public Vector<Double> slopes; 
	// The actual slopes of the codeblocks.
	public Vector<Float> cblkSlopes; 

	public SubbandRateDistStats() {
		rates = new Vector<Double>();
		phbytes = new Vector<Double>();
		dists = new Vector<Double>();
		slopes = new Vector<Double>();
		cblkSlopes = new Vector<Float>();
	}

	// Calculate slopes for Rate-Distortion-Information.
	public void calcSlopes() {
		double prevRate = rates.get(0);
		double prevDist = dists.get(0);
		slopes.add((double) Float.MAX_VALUE);
		for (int i = 1; i < rates.size(); i++) {
			double rate = rates.get(i);
			double dist = dists.get(i);
			double slope = ((dist - prevDist) / (prevRate - rate));
			slopes.add(slope);
			prevDist = dist;
			prevRate = rate;
		}
	}

	public void multDists(double m) {
		for (int i = 0; i < dists.size(); i++) {
			dists.set(i, dists.elementAt(i).doubleValue() * m);
		}
	}

	// Taken from Taubman's JPEG2000 book, p.345.
	public void calcConvexHullDirect() {
		Vector<Double> allSlopes = new Vector<Double>();
		allSlopes.add(new Double(Float.MAX_VALUE));

		for (int z = 1; z < rates.size(); z++) {
			allSlopes.add(new Double(Float.MAX_VALUE));
			for (int t = z - 1; t >= 0; t--) {
				double deltaDist = dists.get(t) - dists.get(z);
				double deltaRate = rates.get(z) - rates.get(t);
				if (deltaDist < allSlopes.get(z) * deltaRate) {
					allSlopes.add(z, deltaDist / deltaRate);
				}
			}
		}

		Vector<Integer> H = new Vector<Integer>();
		for (int z = 0; z < rates.size(); z++) {
			for (int t = z + 1; t < rates.size(); t++) {
				double deltaDist = dists.get(z) - dists.get(t);
				double deltaRate = rates.get(t) - rates.get(z);
				if (deltaDist > allSlopes.get(z) * deltaRate) {
					if (!H.contains(new Integer(z))) {
						break;
					}
				}
			}
			if (allSlopes.get(z) > 0) {
				H.add(z);
			}
		}

		// Copy only points on convex hull.
		copyOnlyPointsOnConvexHull(allSlopes, H);
	}

	private void copyOnlyPointsOnConvexHull(Vector<Double> allSlopes,
			Vector<Integer> H) {
		Vector<Double> nSlopes = new Vector<Double>();
		Vector<Double> nDists = new Vector<Double>();
		Vector<Double> nRates = new Vector<Double>();
		Vector<Float> ncblkSlopes = new Vector<Float>();

		for (int i = 0; i < H.size(); i++) {
			nDists.add(dists.elementAt(H.get(i)));
			nRates.add(rates.elementAt(H.get(i)));
			nSlopes.add(allSlopes.elementAt(H.get(i)));
			ncblkSlopes.add(cblkSlopes.elementAt(H.get(i)));
		}

		dists = nDists;
		rates = nRates;
		slopes = nSlopes;
		cblkSlopes = ncblkSlopes;
	}

	// Convex hull computation as sketched in Taubmans JPEG2000 book, p.346
	// (efficient)
	public void calcConvexHull() {
		Vector<Integer> H = new Vector<Integer>();
		Vector<Double> allSlopes = new Vector<Double>();
		for (int z = 0; z < rates.size(); z++)
			allSlopes.add(new Double(Float.MAX_VALUE));

		H.add(0);
		int hlast = 0;

		for (int z = 1; z < rates.size(); z++) {
			double deltaDist = dists.get(hlast) - dists.get(z);
			double deltaRates = rates.get(z) - rates.get(hlast);
			if (deltaDist > 0) {
				while (deltaDist > allSlopes.get(hlast) * deltaRates) {
					H.removeElement(new Integer(hlast));
					hlast = Collections.max(H);
					deltaDist = dists.get(hlast) - dists.get(z);
					deltaRates = rates.get(z) - rates.get(hlast);
				}
				hlast = z; // + 1; // Error in the book?
				H.add(hlast);
				allSlopes.set(hlast, deltaDist / deltaRates);
			}
		}

		// Copy only points on convex hull.
		copyOnlyPointsOnConvexHull(allSlopes, H);
	}

	public float getCodeBlockRDThreshold(float sbbRDThreshold) {
		int n;
		for (n = 0; n < slopes.size(); n++) {
			if (slopes.elementAt(n) < sbbRDThreshold) {
				break;
			} else
				continue;
		}
		// Store the index in the codBlock.truncIdxs that gives the
		// real truncation index.
		return cblkSlopes.elementAt(n - 1);
	}

	public void check() {
		StringBuffer strBuffer = new StringBuffer(
				"Subband Rate-Distortion Statistics:\n");
		for (int i = 0; i < rates.size(); i++) {
			strBuffer.append("rate = " + rates.get(i) + "\t");
			strBuffer.append("dist = " + dists.get(i) + "\t");
			strBuffer.append("phco = " + phbytes.get(i) + "\t\n");

		}

		double prev = phbytes.get(0);
		for (int i = 1; i < phbytes.size(); i++) {
			double cur = phbytes.get(i);
			if (cur < prev) {
				System.out.println("PacketHeaderNotMonotoneByNBytes: cur = "
						+ cur + ", prev =" + prev);

			}
			prev = cur;
		}

		prev = slopes.get(0);
		for (int i = 1; i < rates.size(); i++) {
			double cur = slopes.get(i);
			if (prev < cur) {
				System.out.println("SlopesNotMonotone: cur = " + cur
						+ ", prev =" + prev);
			}
			prev = cur;
		}
	}

	public String toString() {
		StringBuffer strBuffer = new StringBuffer(
				"Subband Rate-Distortion Statistics:\n");
		for (int i = 0; i < rates.size(); i++) {
			strBuffer.append("rate = " + rates.get(i) + "\t");
			strBuffer.append("dist = " + dists.get(i) + "\t\n");
		}

		strBuffer.append("\n #(Slopes)");
		for (int i = 0; i < slopes.size(); i++) {
			strBuffer.append("\t" + slopes.get(i));
		}
		strBuffer.append("\n");
		return strBuffer.toString();
	}

	// Only for debugging / testing.
	public static void main(String args[]) {
		SubbandRateDistStats sbrds = new SubbandRateDistStats();
		System.out.println(sbrds);
		
		sbrds.rates.add(0.);
		sbrds.dists.add(6.);
		sbrds.rates.add(1.);
		sbrds.dists.add(4.);
		sbrds.rates.add(2.);
		sbrds.dists.add(4.);
		sbrds.calcSlopes();
		System.out.println(sbrds);
		sbrds.calcConvexHull();
		System.out.println(sbrds);
	}
}
