/*
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * */
package jj2000.j2k.entropy.encoder;

import jj2000.j2k.codestream.writer.*;

import jj2000.j2k.wavelet.analysis.*;
import jj2000.j2k.entropy.encoder.*;
import jj2000.j2k.codestream.*;
import jj2000.j2k.entropy.*;
import jj2000.j2k.encoder.*;
import jj2000.j2k.image.*;
import jj2000.j2k.util.*;

import java.util.Collections;
import java.util.Hashtable;
import java.util.Vector;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.*;
import java.lang.reflect.Array;

/**
 * This implements a simple post compression rate allocation algorithm. This
 * algorithm finds the most suitable truncation points for the set of
 * code-blocks, for a final target bitrate. It works by first collecting the
 * rate distortion info from all code-blocks, in all tiles and all components,
 * and then running the rate-allocation on the whole image at once, for each
 * layer.
 * 
 * @see PostCompRateAllocator
 * 
 * @see CodedCBlkDataSrcEnc
 * 
 * @see jj2000.j2k.codestream.writer.CodestreamWriter
 * */
public class WPRateAllocator extends EBCOTRateAllocator {

	private Hashtable subbandRDStats;
	private double mainHeaderQCD_QCC_Overhead = 0; // 2 if expounded, 0 derived,
													// 1 if lossless
	private double mainHeaderCOD_WPS_Overhead = 1 / 8.; // 1 bit has to be
														// written for every
														// subband (decomposed
														// or final).

	public WPRateAllocator(CodedCBlkDataSrcEnc src, LayersInfo lyrs,
			CodestreamWriter writer, EncoderSpecs encSpec, ParameterList pl) {

		super(src, lyrs, writer, encSpec, pl);

		subbandRDStats = new Hashtable();

		// RDO H WPB Specific Initialization
		String qType = (String) encSpec.qts.getDefault();
		if (qType.equals("reversible")) {
			mainHeaderQCD_QCC_Overhead = 1;
		}
		if (qType.equals("derived")) {
			mainHeaderQCD_QCC_Overhead = 0;
		}
		if (qType.equals("expounded")) {
			mainHeaderQCD_QCC_Overhead = 2;
		}
	}

	public void initialize() throws IOException{
		try{
			super.initialize();
		}catch(IllegalArgumentException e){
			// Ignore too low bitrates.
		}
	}
	
	public int getAvailableBytes(){
		float rate = encSpec.pl.getFloatParameter("rate");
        return  (int)(rate * src.getImgWidth() * src.getImgHeight()/8);        
	}
	
	public RDResult getRDCostForSubband(int subbandIdx, float rdThreshold) {
		
		int headerStrategy = encSpec.pl.getIntParameter("Aheader_alg");
		
		if (encSpec.pl.getBooleanParameter("Bdirect")) {
			return getRDCostOfSubbandDirect(subbandIdx, rdThreshold,
					headerStrategy);
		} else {
			return getRDCostForSubbandIndirect(subbandIdx, rdThreshold,
					headerStrategy);
		}
	}

	public RDResult getRDCostForSubbandIndirect(int subbandIdx,
			float rdThreshold, int headerStrategy) {
		// Get SubbandRateDistStats.
		int res = getResLevelFromSubbandIndex(subbandIdx);
		SubbandRateDistStats sbbRD;
		if (subbandRDStats.containsKey(new String(subbandIdx + "-" + res))) {
			sbbRD = (SubbandRateDistStats) subbandRDStats.get(new String(
					subbandIdx + "-" + res));
		} else {
			sbbRD = getRDStatOfSubband(0, 0, subbandIdx, res, headerStrategy);
			subbandRDStats.put(new String(subbandIdx + "-" + res), sbbRD);
		}
		int slopeIdx;
		for (slopeIdx = 0; slopeIdx < sbbRD.slopes.size(); slopeIdx++) {
			float curSlope = (sbbRD.slopes.get(slopeIdx)).floatValue();
			if (curSlope < rdThreshold) {
				break;
			}
		}
		slopeIdx--;
		if (slopeIdx < 0) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,
				"Subband not included. This should never happen.");
		}

		return new RDResult(sbbRD.rates.get(slopeIdx), sbbRD.dists
				.get(slopeIdx), rdThreshold);
	}
	
	/**
	 * Calculates the Rate-Distortion Statistic for a subband. // TODO How to
	 * debug / verify?
	 * 
	 * @param component
	 * @param tile
	 * @param subbandIdx
	 * @param lambda
	 * @return
	 */
	public SubbandRateDistStats getRDStatOfSubband(int tile, int component,
			int subbandIdx, int res, int headerStrategy) {
		// Caching.
		SubbandRateDistStats sbbRD;
		if (subbandRDStats.containsKey(new String(subbandIdx + "-" + res))) {
			sbbRD = (SubbandRateDistStats) subbandRDStats.get(new String(
					subbandIdx + "-" + res));
			return sbbRD;
		}
		sbbRD = new SubbandRateDistStats();

		// Get subband properties.
		int rlev = res;
		SubbandAn sb = src.getSubbandTree(tile, component);
		sb = (SubbandAn) sb.getSubbandByIdx(rlev, subbandIdx);
		Coord ncblks = src.getNumCodeBlocks(sb, null);

		// Get all the codeblocks of the current subband.
		CBlkRateDistStats subcblks[] = cblks[tile][component][rlev][subbandIdx];
		int layer = 0; // Quality layer index, we only consider one layer.
		int precinctIdx = 0; // We only consider one precinct.

		// Get all possible slopes of all codeblocks.
		Vector<Float> allSlopes = new Vector<Float>();

		allSlopes.add(Float.MAX_VALUE); // Consider no contribution as well;
		allSlopes.add(0.f); // Consider all contributions as well;

		for (int b = 0; b < subcblks.length; b++) {
			// Take the current cblk stats.
			CBlkRateDistStats ccb = subcblks[b];
			for (int slopeIdx = 0; slopeIdx < ccb.truncSlopes.length; slopeIdx++) {
				allSlopes.add(ccb.truncSlopes[slopeIdx]);
			}
		}

		// The slopes start with the largest absolute value.
		Collections.sort(allSlopes);
		Collections.reverse(allSlopes);
		// System.out.println("All slopes of a subband " + allSlopes);
		sbbRD.cblkSlopes = allSlopes;
		// For all lambdas (of all codeblocks):
		// Find rate-distortion and add packet header cost to the rate.
		// What do we need for calculating the packet header cost?
		for (int i = 0; i < allSlopes.size(); i++) {
			float rdThreshold = allSlopes.get(i);
			// The cumulative distortion of the subband for the current
			// rdThreshold.
			double cDist = 0;
			// The cumulative rate in bytes of the subband for the current
			// rdThreshold.
			double cRate = 0;

			// Loop on codeblocks and find indices for the current rdThreshold.
			for (int b = 0; b < subcblks.length; b++) {
				// Take the current cblk.
				CBlkRateDistStats ccb = subcblks[b];
				// System.out.println(ccb);
				// Search for the index with a slope more or equal to
				// rdThreshold.
				float curSlope;
				int slopeIdx;
				// Loop on valid slopes
				for (slopeIdx = 0; slopeIdx < ccb.nVldTrunc; slopeIdx++) {
					curSlope = ccb.truncSlopes[slopeIdx];
					if (curSlope < rdThreshold) {
						break;
					}
				}
				slopeIdx--;

				// Save the slopeIdx in array truncIdxs (Necessary for to get
				// the packet header cost)
				truncIdxs[tile][layer][component][rlev][subbandIdx][ccb.m
						* ncblks.x + ccb.n] = slopeIdx;

				if (slopeIdx < 0) {
					// A codeblock that does not contributes generates the
					// maximum distortion, but does not add any bytes to the
					// rate.
					double max_dist_reduction = getMaxOfDoubleArray(ccb.truncDists);
					// Calculate the overall dist of the subband.
					cDist += max_dist_reduction;
					// Add nothing to the overall rate, as this codeblock does
					// not contribute.
					continue;
				}
				// Get actual truncDists- and truncRates-Index.
				int rdIdx = ccb.truncIdxs[slopeIdx];

				// Calculate the cumulative distortion reduction for this index.
				double cdr = ccb.truncDists[rdIdx];
				double max_dist_reduction = getMaxOfDoubleArray(ccb.truncDists);
				double dist = (max_dist_reduction - cdr);

				cDist += dist;

				// Calculate the overall rate of the subband.
				// The used rate in bytes.
				int rate = ccb.truncRates[rdIdx];
				cRate += rate;

			}// End loop on codeblocks of the subband.

			/* begin packet header consideration. */
			if (headerStrategy == 0) {
				// Computation of packet header.
				double subbandPacketHeaderCost = 
					getSubbandPacketHeaderCost(subbandIdx, layer, component, rlev, tile, precinctIdx);
				
				sbbRD.phbytes.add(subbandPacketHeaderCost);
				cRate += subbandPacketHeaderCost;
				cRate += mainHeaderQCD_QCC_Overhead;
				cRate += mainHeaderCOD_WPS_Overhead;

			} else { 
				double estHC = estimateSubbandHeaderCost(cRate, subcblks.length, headerStrategy);
				cRate += estHC;
				sbbRD.phbytes.add(estHC);
			}
			/* end packet header consideration */
			
			// Calculate the overall rate of the subband.
			// Add to Subband Rate distortion stat.
			sbbRD.dists.add(cDist);
			sbbRD.rates.add((double) cRate);
		}
		sbbRD.calcConvexHull();

		packetEnc.reset();
		// Caching.
		subbandRDStats.put(new Integer(subbandIdx), sbbRD);
		return sbbRD;
	}

	public RDResult getRDCostOfSubbandDirect(int subbandIdx, float rdThreshold,
			int headerStrategy) {
		int rlev = getResLevelFromSubbandIndex(subbandIdx);

		// Sum of Lagragian costs
		double cCostJ = 0;
		// The cumulative distortion of the subband for a certain lambda.
		double cDist = 0;
		// The cumulative rate in bytes of the subband for a certain lambda.
		double cRate = 0;

		int t = 0; // tile index
		int c = 0; // component index
		int l = 0; // quality layer index
		int precinctIdx = 0;

		// Get all the codeblocks of the current subband.
		CBlkRateDistStats subcblks[] = cblks[t][c][rlev][subbandIdx];

		// Loop on codeblocks of subband and sum up rate and distortion with
		// slopes more or equal to lambda.
		for (int b = 0; b < subcblks.length; b++) {
			// Take the current cblk.
			CBlkRateDistStats ccb = subcblks[b];

			// Search for the index with a slope more or equal to lambda.
			float curSlope;
			int slopeIdx;
			// Loop on valid slopes
			for (slopeIdx = 0; slopeIdx < ccb.nVldTrunc; slopeIdx++) {
				curSlope = ccb.truncSlopes[slopeIdx];
				if (curSlope < rdThreshold) {
					break;
				}
			}
			slopeIdx--;

			/* begin packet header consideration. */
			// Save it in array truncIdxs (Necessary for to get the packet
			// header cost)
			SubbandAn sb = src.getSubbandTree(t, c);
			sb = (SubbandAn) sb.getSubbandByIdx(rlev, subbandIdx);
			Coord ncblks = src.getNumCodeBlocks(sb, null);
			truncIdxs[t][l][c][rlev][subbandIdx][ccb.m * ncblks.x + ccb.n] = slopeIdx;
			/* end packet header consideration. */

			if (slopeIdx < 0) {
				// A codeblock that does not contributes generates the maximum
				// distortion, but does not add any bytes to the rate.
				double max_dist_reduction = getMaxOfDoubleArray(ccb.truncDists);
				// Calculate the overall dist of the subband.
				cDist += max_dist_reduction;
				// Add nothing to the overall rate, as this codeblock does not
				// contribute.
				continue;
			}
			// Get actual truncDists- and truncRates-Index.
			int rdIdx = ccb.truncIdxs[slopeIdx];

			// Calculate the cummulative distortion.
			// The cumulative distortion reduction for this index.
			double cdr = ccb.truncDists[rdIdx];
			double max_dist_reduction = getMaxOfDoubleArray(ccb.truncDists);
			double dist = (max_dist_reduction - cdr);

			// Calculate the overall dist of the subband.
			cDist += dist;

			// The used rate in bytes.
			int rate = ccb.truncRates[rdIdx];

			// Calculate the overall rate of the subband.
			cRate += rate;

		}
		// Calculate the Lagrangian cost of the subband.
		/* begin packet header consideration. */
		if (headerStrategy == 0) {
			// Computation of packet header.
			double subbandPacketHeaderCost = 
				getSubbandPacketHeaderCost(subbandIdx, l, c, rlev, t, precinctIdx);
			
			cRate += subbandPacketHeaderCost;
			cRate += mainHeaderQCD_QCC_Overhead;
			cRate += mainHeaderCOD_WPS_Overhead;

		} else { 
			double estHC = estimateSubbandHeaderCost(cRate, subcblks.length, headerStrategy);
			cRate += estHC;
		}

		cCostJ += cDist + rdThreshold * cRate;
		return new RDResult(cRate, cDist, rdThreshold);
	}

	private double getSubbandHeaderCost(int subbandIdx, int layer, 
			int component, int rlev, int tile, int precinctIdx){
	
			double headerCost = 				
				getSubbandPacketHeaderCost(subbandIdx, layer, component, rlev, tile, precinctIdx);
			headerCost += mainHeaderQCD_QCC_Overhead;
			headerCost += mainHeaderCOD_WPS_Overhead;
			return headerCost;
	}

	private double getSubbandPacketHeaderCost(int subbandIdx, int layer, 
			int component, int rlev, int tile, int precinctIdx){
		double subbandPacketHeaderCost = 0;
		BitOutputBuffer hBuff = new BitOutputBuffer();
		packetEnc.reset();
		hBuff = packetEnc.encodePacketHeaderOfSubband(subbandIdx,
				layer + 1, component, rlev, tile,
				cblks[tile][component][rlev],
				truncIdxs[tile][layer][component][rlev], hBuff,
				precinctIdx);
	
		subbandPacketHeaderCost = hBuff.getBitLength() / 8.;
		return subbandPacketHeaderCost;
	}

	private double estimateSubbandHeaderCost(double cRate, int nCblks, int headerStrategy) {
		
		double estHC = 0;
		if (headerStrategy == 1) {
			// No consideration of header data.
			return 0.;
		} else if (headerStrategy == 2) {
			// Estimate packet header.
			
			if (cRate == 0) {
				estHC = 1;
			} else{
				if (nCblks == 1) {
					estHC = 1 + mainHeaderQCD_QCC_Overhead + mainHeaderCOD_WPS_Overhead;
				} else {
					estHC = 4 + mainHeaderQCD_QCC_Overhead + mainHeaderCOD_WPS_Overhead;
				}
			}
			return estHC;
		}
		// Default: return estimate.
		return estHC;
	}

	public int getMainHeaderLength() {
		int mainHeaderLength = 0;
		try {
			headEnc.reset();
			headEnc.encodeMainHeader(getNumLayers());
			mainHeaderLength = headEnc.getLength();
		} catch (Exception e) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING,
				"RDOWPBRateAllocator.getMainHeaderLength(): Not yet initialized!");
		}
		
		return mainHeaderLength;
	}

	public int getTilePartHeaderLength(int tile, int component) {
		int tilePartHeaderLength = 0;
		try {
			headEnc.reset();
			headEnc.encodeTilePartHeader(tile, component, 1);
			tilePartHeaderLength = headEnc.getLength();
		} catch (Exception e) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING,
				"RDOWPBRateAllocator.getTilePartHeaderLength(): Not yet initialized!");
		}
		return tilePartHeaderLength;
	}

	public void runAndWrite() throws IOException, IllegalArgumentException {

		if (encSpec.pl.getParameter("Asb_thresh") != null) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
				"Asb_thresh set. Using WPRateAllocator.write().");
			float sbbRDThreshold = encSpec.pl.getFloatParameter("Asb_thresh");
			write(sbbRDThreshold, true);
		} else if (encSpec.pl.getParameter("Athresh") != null) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
				"Athresh set. Using WPRateAllocator.write().");
			float RDThreshold = encSpec.pl.getFloatParameter("Athresh");
			write(RDThreshold, false);
		} else {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
				"Using EBCOTRateAllocator.runAndWrite().");
			super.runAndWrite();
		}
	}
	
	/**
	 * Write a simple codestream, with the given RDThreshold; which is either for 
	 * subbands or for codeblocks rate distortion statistics.
	 * 
	 * @param RDThreshold for the rate distortion optimization.
	 * @param sbbRD if true subband rate distortion statistics are used else 
	 * 				codeblock rate distortion statistics are used.
	 * @throws IOException
	 * @throws IllegalArgumentException
	 */
	protected void write(float RDThreshold, boolean sbbRD) throws IOException,
			IllegalArgumentException {
		int numComps = src.getNumComps();
		int numTiles = src.getNumTiles();
		int[][] mrl = packetEnc.getMRL();
		BitOutputBuffer hBuff = null;
		byte[] bBuff = null;

		if (numComps > 1 || numTiles > 1) {
			System.out
					.println("Multiple components and tiles are not supported.");
			throw new IllegalArgumentException();
		}

		int t = 0;
		int c = 0;
		int l = 0;
		int precinctIdx = 0;
		int maxRes = mrl[t][c];

		// Tile header
		headEnc.reset();
		// According to the standard a value of 0 is permitted (for last tile).
		headEnc.encodeTilePartHeader(0, t, numLayers); 
		bsWriter.commitBitstreamHeader(headEnc);

		// Should SOP markers be used ?
		boolean sopUsed = ((String) encSpec.sops.getTileDef(t)).equals("on"); 
		// Should EPH markers be used ?
		boolean ephUsed = ((String) encSpec.ephs.getTileDef(t)).equals("on"); 

		// Loop on resolutions
		for (int r = 0; r <= maxRes; r++) {

			setTruncIdxs(r, RDThreshold, sbbRD);

			packetEnc.reset();

			hBuff = packetEnc.encodePacket(l + 1, c, r, t, cblks[t][c][r],
					truncIdxs[t][l][c][r], hBuff, bBuff, precinctIdx);

			if (packetEnc.isPacketWritable()) {
				writeHeaderAndBody(false, hBuff, sopUsed, ephUsed);
			}
		}
	}

	
	protected void setTruncIdxs(int lvlIdx, float RDThreshhold, boolean bSbbRD) {
		int tileIdx = 0;
		int compIdx = 0;
		int layerIdx = 0;

		int precinctIdx = 0;
		int headerStrategy = 0;

		// write packet header and body
		int s, b, bIdx, n;
		SubbandAn sb;
		CBlkRateDistStats cur_cblk;
		Vector cbvec = null;

		// Loop on subbands of resolutions
		sb = (SubbandAn) src.getSubbandTree(tileIdx, compIdx)
				.getLowestLeafInRes(lvlIdx);
		while (sb != null) { // loop on subbands

			// Get subband RDStats
			// Get RD threshold for codeblocks
			// Get truncIdxs
			s = sb.sbandIdx;
			
			float fthresh;
			if (bSbbRD) {
				SubbandRateDistStats sbbRDS = getRDStatOfSubband(tileIdx, 
						compIdx, s, lvlIdx, headerStrategy);
				fthresh = sbbRDS.getCodeBlockRDThreshold(RDThreshhold);
			} else {
				fthresh = RDThreshhold;
			}
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
					"Fthresh for codeblocks = " + fthresh);

			cbvec = packetEnc.getCBlkInPrec(tileIdx, compIdx, lvlIdx, s,
					precinctIdx, cbvec);
			
			for (bIdx = 1; bIdx < cbvec.size(); bIdx++) {
				// Get the current code-block
				b = ((Integer) cbvec.elementAt(bIdx)).intValue();
				cur_cblk = cblks[tileIdx][compIdx][lvlIdx][s][b];
				// System.out.println(cur_cblk);
				for (n = 0; n < cur_cblk.nVldTrunc; n++) {
					if (cur_cblk.truncSlopes[n] < fthresh) {
						break;
					} else
						continue;
				}
				// Store the index in the codBlock.truncIdxs that gives the
				// real truncation index.
				truncIdxs[tileIdx][layerIdx][compIdx][lvlIdx][s][b] = n - 1;
			}
			sb = (SubbandAn) sb.nextSubband();
		}
	}
	
	public double getMaxOfDoubleArray(double[] t) {
		if (t.length == 0)
			return 0;
		double maximum = t[0];
		for (int i = 1; i < t.length; i++) {
			if (t[i] > maximum) {
				maximum = t[i]; // new maximum
			}
		}
		return maximum;
	}

	/**
	 * Returns the resolution level of a subband with subbandIdx. Only works for
	 * full wavelet packet decompositions.
	 * 
	 * @param subbandIdx
	 *            the subband index of a fully decomposed wavelet tree
	 * @return the resolution level of the subband
	 */
	public static int getResLevelFromSubbandIndex(int subbandIdx) {
	
		for (int i = 0;; i++)
			if (subbandIdx < Math.pow(4, i))
				return i;
	}
}