/*
 * CVS identifier:
 *
 * $Id: ArrayUtil.java,v 1.7 2006-05-22 08:21:11 dengel Exp $
 *
 * Class:                   ArrayUtil
 *
 * Description:             Utillities for arrays.
 *
 *
 *
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */


package jj2000.j2k.util;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;

/**
 * This class contains a colleaction of utility static methods for arrays.
 * */
public class ArrayUtil {

    /** The maximum array size to do element by element copying, larger
     * arrays are copyied in a n optimized way. */
    public static final int MAX_EL_COPYING = 8;

    /** The number of elements to copy initially in an optimized array copy */
    public static final int INIT_EL_COPYING = 4;

    /**
     * Reinitializes an int array to the given value in an optimized way. If
     * the length of the array is less than MAX_EL_COPYING, then the array
     * is set element by element in the normal way, otherwise the first
     * INIT_EL_COPYING elements are set element by element and then
     * System.arraycopy is used to set the other parts of the array.
     *
     * @param arr The array to set.
     *
     * @param val The value to set the array to.
     *
     *
     * */
    public static void intArraySet(int arr[], int val) {
        int i,len,len2;
        
        len = arr.length;
        // Set array to 'val' in an optimized way
        if (len < MAX_EL_COPYING) { 
            // Not worth doing optimized way
            for (i=len-1; i>=0; i--) { // Set elements
                arr[i] = val;
            }
        }
        else { // Do in optimized way
            len2 = len>>1;
            for (i=0; i<INIT_EL_COPYING; i++) { // Set first elements
                arr[i] = val;
            }
            for (; i <= len2 ; i<<=1) {
                // Copy values doubling size each time
                System.arraycopy(arr,0,arr,i,i);
            }
            if (i < len) { // Copy values to end
                System.arraycopy(arr,0,arr,i,len-i);
            }
        }
    }

    /**
     * Reinitializes a byte array to the given value in an optimized way. If
     * the length of the array is less than MAX_EL_COPYING, then the array
     * is set element by element in the normal way, otherwise the first
     * INIT_EL_COPYING elements are set element by element and then
     * System.arraycopy is used to set the other parts of the array.
     *
     * @param arr The array to set.
     *
     * @param val The value to set the array to.
     *
     *
     * */
    public static void byteArraySet(byte arr[], byte val) {
        int i,len,len2;
        
        len = arr.length;
        // Set array to 'val' in an optimized way
        if (len < MAX_EL_COPYING) { 
            // Not worth doing optimized way
            for (i=len-1; i>=0; i--) { // Set elements
                arr[i] = val;
            }
        }
        else { // Do in optimized way
            len2 = len>>1;
            for (i=0; i<INIT_EL_COPYING; i++) { // Set first elements
                arr[i] = val;
            }
            for (; i <= len2 ; i<<=1) {
                // Copy values doubling size each time
                System.arraycopy(arr,0,arr,i,i);
            }
            if (i < len) { // Copy values to end
                System.arraycopy(arr,0,arr,i,len-i);
            }
        }
    }
    
    /**
     * print a _byte_ array as hex sequence 
     * (note: for wp structures use bit arrays,
     * which look the same, but are interpreted
     * slightly differently)
     *
     */
    public static String printByteArrayAsHex(byte buf [])
    {
    	String tmp = "";
    	for (int i = 0; i < buf.length; i++)
    	{
    		int val = 0;
    		val = 0xff & buf[i];
    		if (val < 16)
    		{
    			tmp += "0";
    		}
    		tmp += Integer.toHexString(val);
    		tmp += " ";
    		
    		if ( (i+1) % 16 == 0) tmp += "\n";    			    	
    	}
    	return tmp;
    
    }
    
    /**
     * convert a bitArray (wpStructure) to a binary String
     * 
     * @param buf
     * @return buf as binary String
     */
    public static String printBitArray (byte buf[])
	{
    	String tmp = "";
    	
       	for (int i=0; i < buf.length; i++)
       	{
    		for (int j=0; j<8; j++)
    		{   			
    			int hlp = 0xff & buf[i];
    			if ((hlp & 1<<j) != 0)
    			{
    				tmp += "1";
    			}
    			else
    			{
    				tmp+= "0";
    			}    			
    		}
       	}
       	return tmp;
	}
    
    /**
     * convert a bitArray (wpStructure) to a hex string
     * 
     * (this is interpreted differently than a byte array)
     * 
     * @param buf
     * @return buf as hex string
     */
    public static String printBitArrayAsHex (byte buf[])
    {
       	String tmp = "";
    	
       	for (int i=0; i < buf.length; i++)
       	{
       		int num = 0;
    		for (int j=0; j<4; j++)
    		{   			
    			int hlp = 0xff & buf[i];    			
    			if ((hlp & 1<<j) != 0)
    			{
    				num += Math.pow(2, j);
    			}
    		}
    		tmp += Integer.toHexString(num);
    		
    		num = 0;
    		for (int j=4; j<8; j++)
    		{
    			int hlp = 0xff & buf[i];    			
    			if ((hlp & 1<<j) != 0)
    			{
    				num += Math.pow(2, j-4);
    			}    			
    		}
    		tmp += Integer.toHexString(num);
    		
       	}
       	return tmp;
    }
    
    
	/**
	 * writes a 1 to buf at bit position bitpos
	 * @param buf the byte buffer
	 * @param bitpos the bitposition
	 */
    public static void writeBit(byte[] buf, int bitpos)
	{
		int pos = bitpos/8;
		int relbitpos = bitpos % 8;
		int hlp;			
		
		if (pos > buf.length -1) throw new IllegalArgumentException();
		
		hlp = 0xFF & buf[pos];		
		buf[pos] =  (byte) (hlp | 1<<relbitpos);					
	}
    
	/**
	 * reads a bit from buf at bit position bitpos
	 * @param buf the byte buffer
	 * @param bitpos the bitposition
	 */
    public static boolean readBit(byte[] buf, int bitpos)
	{
		int pos = bitpos/8;
		int relbitpos = bitpos % 8;
		int hlp;			
		
		if (pos > buf.length -1) 
			{
				FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "Could not read requested bit from Bitarray.");
				return false;
				//throw new IllegalArgumentException();
			}
		
		hlp = 0xFF & buf[pos];		
		
		return  ((hlp &  (1<<relbitpos)) != 0)  ? true : false;
	}
    
    
    /**
     * Create bit array from a bit String or a hex String
     * @param fname
     * @return
     */
    public static byte[] createBitArrayFromString(String wps) 
    {
    		if (wps == "" ) return null;
    		
    		if (wps.substring(0,1).equalsIgnoreCase("b"))
    		{
    			for(int i=1; i<wps.length(); i++)
    				if(wps.charAt(i) !='0' && wps.charAt(i) !='1') {
    		    		FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "WPS:" + wps+" is not a Bit tring and is interpreted as HEX string");
    	    			return createBitArrayFromHexString(wps);
   					
    				}
    					
    			return createBitArrayFromBitString(wps.substring(1));
    		}
    		else
    		{
    			return createBitArrayFromHexString(wps);
    		}
    	
    }
    
    /**
     * Create a bit array from a string
     * 
     * @param arrString
     * @return bitarray corresponding to arrString as byte array
     */
    public static byte[] createBitArrayFromBitString (String arrString)
    {
    	if (arrString == "") return null;
    	
    	byte[] buf = new byte[arrString.length()/8 + 1];
    	
    	for (int bitpos =0; bitpos < arrString.length(); bitpos++)
    	{    		
    		if (arrString.substring(bitpos, bitpos+1).equals("1"))
    		{
    			writeBit(buf, bitpos);
    		}
    		else if (arrString.substring(bitpos, bitpos+1).equals("0"))
    		{
    			//just increment
    		}
    		else
    		{
    			//FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "Malformed WPS String: " + arrString);
    			throw new IllegalArgumentException("Malformed WPS String " + arrString);
    		}
    	}    	    	
    	return buf;    	
    }
    
    /**
     * Create a bit array from a string
     * 
     * @param arrString
     * @return bitarray corresponding to arrString as byte array
     */
    public static byte[] createBitArrayFromHexString (String hexString)
    {
    	
    	// WSQ = f101248124803480148010
    	
    	if (hexString == "") return null;
    	
    	if (hexString.substring(0,2).equalsIgnoreCase("0x")) //also allow initial 0x
    	{
    		hexString = hexString.substring(2);
    	}
    	    	
    	byte[] buf = new byte[(hexString.length()+1)/2]; //two hexes per byte
    	
    	for (int hexpos =0; hexpos < hexString.length(); hexpos++)
    	{
    		int tmp;    
    		tmp = Integer.parseInt(hexString.substring(hexpos, hexpos+1), 16);
    		
    		for (int bitpos =0; bitpos < 4; bitpos++)
	    	{    		
	    		if ( (tmp & (1<<bitpos)) != 0 )
	    		{
	    			writeBit(buf, 4*hexpos + bitpos);
	    		}
	    	}
    	}
	    return buf;    	
    }
    

    
    /**
     * read a hex string that specifies a bit array from fname
     * 
     * @param fname
     * @return bit array as byte[]
     * @throws IOException
     */
    public static  byte[] readBitArrayFromHexFile (String fname) throws IOException
	{
    	
    	byte[] result = null;
    	String wps = "";
    	
    	try 
		{
    		File f = new File(fname);
    		FileReader fr = new FileReader(f);
    		char[] cbuf = new char [(int) f.length()];
    		fr.read(cbuf);
        	wps = String.valueOf(cbuf);
        	result = createBitArrayFromHexString(wps);
        	fr.close();
		}
    	catch (IOException ioe)
		{
    		FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "could not read wps string");
    		throw ioe;
		}
    	
    	return result;
	}
    
    
    /**
     * write a bit array to a file as a hex string
     * 
     * @param fname target file name
     * @param bits bit array to write
     * @throws IOException
     */
    public static void writeBitArrayToHexFile (String fname, byte[] bits) throws IOException 
	{
		try
		{			
			FileOutputStream fos = new FileOutputStream (fname, false);
			DataOutputStream out = new DataOutputStream(fos);
			out.writeBytes(printBitArrayAsHex(bits));		
			out.close();
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "File not Found"+fnfe.getMessage());
		}
		catch (IOException ioe) {			
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "IO Error");	
		}
	}
    public static String getWlev(String swps) {
		byte[] wps = createBitArrayFromString(swps);
	
		int bp =0;
		while (	ArrayUtil.readBit(wps, bp)) {
			bp++;
		}
		return Integer.toString(bp);
    }

 
    
     
    
}
