/* 
 * CVS identifier:
 * 
 * $Id: ParamKeyUtil.java,v 1.4 2005-11-15 13:58:34 dengel Exp $
 * 
 * Class:                   ParamKeyUtil
 * 
 * Description:             
 * 
 * Created on Dec 14, 2004
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */package jj2000.j2k.util.param;

import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;

/**
 * 
 * Utilities for manipulating key specifications for encryption with parameterized filter transform
 * @author dengel
 *
 */
public abstract class ParamKeyUtil {

	/**
	 * the number of digits that make up a single parameter value
	 */
	 private final static int val_size = 2;
	 
	 /**
	  * create a filterstring from the given key (hex or int)
	  * 
	  * @param filter_prefix
	  * @param key
	  * @return
	  */
	 public static String createFilterStringFromKey (String filter_prefix, String key)
	 {
	 	if (key.charAt(0) == 'x')
	 	{
	 		return createFilterStringFromHexKey(filter_prefix, key.substring(1));
	 	}
	 	else
	 	{
	 		return createFilterStringFromIntKey(filter_prefix, key);
	 	}	 		
	   
	 }
	 
		/**
		 * 
		 * translate an integer key string to a filter string
		 * 
		 * @param key a string containing the horizontal and vertical parameters for each decomposition level
		 * @return filter string corresponding to key
		 */
		private static String createFilterStringFromIntKey (String filter_prefix, String key)
		{

			String res = "";
			int key_val;
			int radix =  (int) (Math.pow(10, val_size)- 1); 
			float alpha;		
			
			if (key.length()/val_size != 10) //assuming decomposition level 5
			{
				FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "No full key specified (" + key + ") [Assuming decomposition level 5]");			
			}
						
			
			for (int i = 0; i < key.length()/2; i+=val_size)
			{				
				alpha=denormalize(Integer.parseInt(key.substring(i, i+val_size)), radix);
				res += filter_prefix + alpha + ":";
			}
			res = res.substring(0, res.length() - 1 ) + ",";
			
			for (int i = key.length()/2; i < key.length(); i+=val_size)
			{				
				alpha=denormalize(Integer.parseInt(key.substring(i, i+val_size)), radix);
				res += filter_prefix + alpha + ":";
			}
			res = res.substring(0, res.length() - 1);
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Converted key " + key + " to filter spec " + res);
							
			return res;		
		}
		
	 
	 
	/**
	 * 
	 * translate a key hexstring to a filter string
	 * 
	 * @param key a string containing the horizontal and vertical parameters for each decomposition level
	 * @return filter string corresponding to key
	 */
	private static String createFilterStringFromHexKey (String filter_prefix, String key)
	{

		String res = "";
		int key_val;
		int radix =  (int) (Math.pow(16, val_size)- 1); 
		float alpha;		
		
		if (key.length()/val_size != 10) //assuming decomposition level 5
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "No full key specified (" + key + ") [Assuming decomposition level 5]");			
		}
					
		
		for (int i = 0; i < key.length()/2; i+=val_size)
		{				
			alpha=denormalize(Integer.parseInt(key.substring(i, i+val_size), 16), radix);
			res += filter_prefix + alpha + ":";
		}
		res = res.substring(0, res.length() - 1 ) + ",";
		
		for (int i = key.length()/2; i < key.length(); i+=val_size)
		{				
			alpha=denormalize(Integer.parseInt(key.substring(i, i+val_size), 16), radix);
			res += filter_prefix + alpha + ":";
		}
		res = res.substring(0, res.length() - 1);
		FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Converted key " + key + " to filter spec " + res);
						
		return res;		
	}
	
	/**
	 * converts the given hex-key to a double 
	 * 
	 * @param key
	 * @return double corresponding to key
	 */
	public static double keyToDouble (String key)
	{
		double val = 0;
		
		for (int i = 0; i < key.length(); i++)
		{
			int tmp = Integer.parseInt(key.substring(key.length()-1 -i, key.length()-i), 16);
			val += tmp * Math.pow (16, i); 
		}
		return val;
		
	}
	

	
	
	/**
	 * Linear denormalize hex value. omit -1.4..0.2 as values for alpha
	 * 
	 * NOTE: Randomized parameterized filters are denormalized by randomGenerator!
	 *  	 		
	 * @param key_val byte value
	 * @return value for alpha
	 */
	protected static float denormalize (int key_val, int radix)
	{		
		
		//   4.6 (-6   .. -1.4) 
		//   4.6 ( 1.4 ..  6  )
		//   ----
		//   9.2 --> step_size = 0.04
		
		System.out.println ("denormalize\n");
		float val = (float) key_val / radix  * 9.2f;
		if (val < 4.6f)
		{
			return ( val - 6f); // --> -6 .. -1.4
		}
		else
		{
			return (val - 3.2f);// --> 1.4 .. 6
		}		
		
	}
}
