/* 
 * CVS identifier:
 * 
 * $Id: WPRandomStructureGenerator.java,v 1.5 2006-10-02 14:02:51 dengel Exp $
 * 
 * Class:                   WPRandomStructureGenerator
 * 
 * Description:             
 * 
 * Created on Mar 8, 2005
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */

package jj2000.j2k.wavelet;

import java.util.Random;
import java.util.StringTokenizer;

import jj2000.j2k.image.BlkImgDataSrc;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;

/**
 * 
 * This class generates random wavelet packet decompositions, based on 
 * a set of parameters. The procedure is described in [Pommer03a].
 * 
 *
 */
public class WPRandomStructureGenerator {

	
	public int max_decomp_global;
	public int max_decomp_approx;
	public int min_decomp_approx;
	public float base_value;
	public float change_factor;
	public int pyramidal_level = 0;
	public long seed;	
	
	private BlkImgDataSrc src;
	
	private Random randomGen;
	

	/**
	 * create random structure generator without imgsrc 
	 * this will only create a subband tree, but the subbands 
	 * will not have a height or a width
	 * 
	 * @param nt
	 * @param nc
	 * @param type
	 */
	public WPRandomStructureGenerator(String spec) {		
		parseSpec(spec);		
	}

	
	/**
	 * 
	 * create random structure generator with imgsrc
	 * use this to create a subband tree and give each 
	 * subband its approriate height and widht, 
	 * based on imgsrc
	 * 
	 * @param nt
	 * @param nc
	 * @param type
	 */
	public WPRandomStructureGenerator(String spec, BlkImgDataSrc imgsrc) {
		this.src = imgsrc;		
		parseSpec(spec);		
	}
	
	/**
	 * create a (pseudo)random wavelet packet decomposition structure
	 * @return
	 */
	public SubbandStructure createRandomStructure()
	{
		SubbandStructure sbbs;
		if (src != null) 
			sbbs = new SubbandStructure(src.getImgWidth(), src.getImgHeight());
		else
			sbbs = new SubbandStructure();
		
		randomGen = new Random (seed);
		
		createRandomStructureRecursive(sbbs);
		
		Subband sb = sbbs.getDeepestGlobalLL();
		if (sb.level != sbbs.resLvl)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Random WPS set new resolution level to: " + sb.level);
			sbbs.setResolutionLevelRecursive(sb.level);
		}
		return sbbs;
	}
	
	private void createRandomStructureRecursive(Subband sb)
	{
		if (sb.level == max_decomp_global) {
			return;
		}

		if (sb.gOrient == Subband.WT_ORIENT_LL) {
			//Approximation Subband
			if (sb.level < min_decomp_approx) {
				decompose(sb);
				return;
			} else if (sb.level == max_decomp_approx) {
				return;
			} else {
				if (inner_decision(sb)) {
					decompose(sb);
					return;
				} else {
					return;
				}
			}
		}
		else {
			//Detail Subband
			if (inner_decision(sb))
				decompose(sb);
			else
				return;
		}
	}
	
	private void  decompose(Subband sb)
	{
		sb.split(null, null);
		if (sb.gOrient == Subband.WT_ORIENT_LL && 
				(sb.level >= (max_decomp_global -  pyramidal_level))) {			
			createRandomStructureRecursive(sb.getLL());
		} else {
			createRandomStructureRecursive(sb.getLL());
			createRandomStructureRecursive(sb.getHL());
			createRandomStructureRecursive(sb.getLH());
			createRandomStructureRecursive(sb.getHH());
		}
	}
	
	private boolean inner_decision (Subband sb)
	{
		float val = randomGen.nextFloat() * 2f;
		float weight = base_value + sb.level * change_factor;
		return ( val / weight >= 1);
	}
	
	
	
	private void parseSpec(String spec)
	{
		StringTokenizer t = new StringTokenizer(spec, "scbgmnp", true);
		
		while (t.hasMoreTokens())
		{
			String token = t.nextToken();
			char var = token.charAt(0);
			switch (var) {
			case 's':
				String sseed = t.nextToken();
				if (sseed.equalsIgnoreCase("t"))
					seed = System.currentTimeMillis();
				else
					seed = Long.parseLong(sseed);								
				break;
			case 'c':
				change_factor = Float.parseFloat(t.nextToken());
				break;
			case 'b':
				base_value = Float.parseFloat(t.nextToken());
				break;
			case 'g':
				max_decomp_global = Integer.parseInt(t.nextToken());
				break;
			case 'm':
				max_decomp_approx = Integer.parseInt(t.nextToken());
				break;
			case 'n':
				min_decomp_approx = Integer.parseInt(t.nextToken());
				break;
			case 'p':
				pyramidal_level = Integer.parseInt(t.nextToken());
				break;
			default:
				throw new IllegalArgumentException("Malformed specification string for WwpsRandom: " + spec);
			}
		}
		FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Random Wavelet Structure Parameters: " + this.toString() );
	}
	
	

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "[WPRandomStructureGenerator: Seed=" + seed + " BaseVal="
				+ base_value + " ChangeFactor=" + change_factor
				+ " MaxGlobalDepth=" + max_decomp_global + " MaxApproxDepth="
				+ max_decomp_approx + " MinApproxDepth=" + min_decomp_approx + "]";
	}
}
