/* 
 * CVS identifier:
 * 
 * $Id: SynWTFilterSpec.java,v 1.10 2006-03-23 10:16:40 dengel Exp $
 * 
 * Class:                   SynWTFilterSpec
 * 
 * Description:             Synthesis filters specification
 * 
 * 
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */
package jj2000.j2k.wavelet.synthesis;

import jj2000.j2k.util.*;
import jj2000.j2k.util.param.ParamKeyUtil;
import jj2000.j2k.util.param.RandomGenerator;
import jj2000.j2k.wavelet.analysis.AnWTFilter;
import jj2000.j2k.wavelet.analysis.AnWTFilterFloatLift9x7;
import jj2000.j2k.wavelet.analysis.AnWTFilterIntLift5x3;
import jj2000.j2k.wavelet.analysis.AnWTParamFilterFloatHaar;
import jj2000.j2k.wavelet.analysis.AnWTParamFilterFloatLift9x7;
import jj2000.j2k.wavelet.analysis.AnWTParamFilterFloatLiftRandom9x7Factory;
import jj2000.j2k.*;

import java.util.*;

/**
 * This class extends ModuleSpec class for synthesis filters
 * specification holding purpose.
 *
 * @see ModuleSpec
 *
 * */
public class SynWTFilterSpec extends ModuleSpec {

    /** The random generator for parameterized filter sequences */
    private RandomGenerator randomGen = null;
    
    /** The paramterlist associated with this object */
    private ParameterList plist = null;
    
	
    /**
     * Constructs a new 'SynWTFilterSpec' for the specified number of
     * components and tiles.
     *
     * @param nt The number of tiles
     *
     * @param nc The number of components
     *
     * @param type the type of the specification module i.e. tile specific,
     * component specific or both.
     *
     * */
    public SynWTFilterSpec(int nt, int nc, byte type){
        super(nt, nc, type);
    }
    
   /**
    * Read Filter Specification from Parameterlist
    * 
    * Used for overriding the filters specified in the bitstream
    * 
    * @param pl Parameterlist containing parameter Ffilters
    */
    public void getSpecFromPL (ParameterList pl){
        
    	this.plist = pl;
    	
        // Check parameters
        pl.checkList(SynWTFilter.OPT_PREFIX,
		     pl.toNameArray(SynWTFilter.getParameterInfo()));	       
        
        if (pl.containsKey("Ffilters") && pl.containsKey("Ffilters_key"))
        {
        	throw new IllegalArgumentException("Ffilters and Ffilters_key are mutually exclusive.");
        }
        
        // test for the seed for the random generator
        if (pl.containsKey("Ffilters_random"))
        {
        	randomGen = new RandomGenerator(pl.getParameter("Ffilters_random"));
        }        
        
    	String param = "";
    	
    	if (pl.containsKey("Ffilters_key"))
    	{
    		param = ParamKeyUtil.createFilterStringFromKey("w9x7p", pl.getParameter("Ffilters_key"));
    	}
    	else
    	{
    		param = pl.getParameter("Ffilters");
    	}


        
    	// No parameter specified, do nothing, filter will be read from bistream
    	if(param==null)  {
    		throw new IllegalArgumentException ("parameterlist for override does not contain option Ffilters");    		
    	}
    	
    	
    	// Parse argument
    	StringTokenizer stk = new StringTokenizer(param);
    	String word; // current word
    	byte curSpecType = SPEC_DEF; // Specification type of the
    	// current parameter
    	boolean[] tileSpec = null; // Tiles concerned by the specification
    	boolean[] compSpec = null; // Components concerned by the specification
    	SynWTFilter[][] filter;
    	
    	while(stk.hasMoreTokens()){
    	    word = stk.nextToken();
    	  
    	    switch(word.charAt(0)){
	    	    case 't': // Tiles specification
	    	    case 'T': // Tiles specification
		     		tileSpec = parseIdx(word,nTiles);
		    		if(curSpecType==SPEC_COMP_DEF)
		    		    curSpecType = SPEC_TILE_COMP;
		    		else
		    		    curSpecType = SPEC_TILE_DEF;
		     		break;
	    	    case 'c': // Components specification
	    	    case 'C': // Components specification
		    		compSpec = parseIdx(word,nComp);
		    		if(curSpecType==SPEC_TILE_DEF)
		    		    curSpecType = SPEC_TILE_COMP;
		    		else
		    		    curSpecType = SPEC_COMP_DEF;
		    		break;
	    	    case 'w': // WT filters specification
	    	    case 'W': // WT filters specification
		    		filter = parseFilters(word);
		    		if(curSpecType==SPEC_DEF){
		    		    setDefault(filter);
		    		}
		    		else if(curSpecType==SPEC_TILE_DEF){
		    		    for(int i=tileSpec.length-1; i>=0; i--)
		    			if(tileSpec[i]){
		    			    setTileDef(i,filter);
		                            }
		    		}
		    		else if(curSpecType==SPEC_COMP_DEF){
		    		    for(int i=compSpec.length-1; i>=0; i--)
		    			if(compSpec[i]){
		    			    setCompDef(i,filter);
		                            }
		    		}
		    		else{
		    		    for(int i=tileSpec.length-1; i>=0; i--){
		    			for(int j=compSpec.length-1; j>=0 ; j--){
		    			    if(tileSpec[i] && compSpec[j]){
		    				setTileCompVal(i,j,filter);
		                                }
		    			}
		    		    }
		    		}
		
		    		// Re-initialize
		    		curSpecType = SPEC_DEF;
		    		tileSpec = null;
		    		compSpec = null;
		    		break;
	
	    	    default:
	    	    	throw new IllegalArgumentException("Bad construction for "+
	    						   	"parameter: "+word); 
	    	    }
    	}
    	

    }


    /**
     * Returns the data type used by the filters in this object, as defined in 
     * the 'DataBlk' interface for specified tile-component.
     *
     * @param t Tile index
     *
     * @param c Component index
     *
     * @return The data type of the filters in this object
     *
     * @see jj2000.j2k.image.DataBlk
     *
     * */
    public int getWTDataType(int t,int c){
	SynWTFilter[][] an = (SynWTFilter[][])getSpec(t,c);
	return an[0][0].getDataType();
    }

    /**
     * Returns the horizontal analysis filters to be used in component 'n' and 
     * tile 't'.
     *
     * <P>The horizontal analysis filters are returned in an array of
     * SynWTFilter. Each element contains the horizontal filter for each
     * resolution level starting with resolution level 1 (i.e. the analysis
     * filter to go from resolution level 1 to resolution level 0). If there
     * are less elements than the maximum resolution level, then the last
     * element is assumed to be repeated.
     *
     * @param t The tile index, in raster scan order
     *
     * @param c The component index.
     *
     * @return The array of horizontal analysis filters for component 'n' and
     * tile 't'.
     *
     *
     * */
    public SynWTFilter[] getHFilters(int t, int c) {
	SynWTFilter[][] an = (SynWTFilter[][])getSpec(t,c);
	return an[0];
    }

    /**
     * Returns the vertical analysis filters to be used in component 'n' and 
     * tile 't'.
     *
     * <P>The vertical analysis filters are returned in an array of
     * SynWTFilter. Each element contains the vertical filter for each
     * resolution level starting with resolution level 1 (i.e. the analysis
     * filter to go from resolution level 1 to resolution level 0). If there
     * are less elements than the maximum resolution level, then the last
     * element is assumed to be repeated.
     *
     * @param t The tile index, in raster scan order
     *
     * @param c The component index.
     *
     * @return The array of horizontal analysis filters for component 'n' and
     * tile 't'.
     *
     *
     * */
    public SynWTFilter[] getVFilters(int t,int c) {
	SynWTFilter[][] an = (SynWTFilter[][])getSpec(t,c);
	return an[1];
    }

    /** Debugging method */
    public String toString(){
	String str = "";
	SynWTFilter[][] an;

	str += "nTiles="+nTiles+"\nnComp="+nComp+"\n\n";

	for(int t=0; t<nTiles; t++){
	    for(int c=0; c<nComp; c++){
		an = (SynWTFilter[][])getSpec(t,c);
		
		str += "(t:"+t+",c:"+c+")\n";
		
		// Horizontal filters
		str += "\tH:";
		for(int i=0; i<an[0].length; i++)
		    str += " "+an[0][i];
		// Horizontal filters
		str += "\n\tV:";
		for(int i=0; i<an[1].length; i++)
		    str += " "+an[1][i];
		str += "\n";
	    }
	}
	
	return str;
    }

    /**
     * Check the reversibility of filters contained is the given
     * tile-component.
     *
     * @param t The index of the tile
     *
     * @param c The index of the component
     *
     */
    public boolean isReversible(int t,int c){
	// Note: no need to buffer the result since this method is
	// normally called once per tile-component.
	SynWTFilter[] 
	    hfilter = getHFilters(t,c),
	    vfilter = getVFilters(t,c);

	// As soon as a filter is not reversible, false can be returned
	for(int i=hfilter.length-1; i>=0; i--)
	    if(!hfilter[i].isReversible() || !vfilter[i].isReversible())
		return false;
	return true;
    }
    
    /** 
     * Parse filters from the given word, with the option to specify an additional
     * parameter for parameterized filters
     *
     * @param word String to parse
     *
     *
     * @return Synthesis wavelet filter (first dimension: by direction,
     * second dimension: by decomposition levels)
     */
    private SynWTFilter[][] parseFilters(String word){
	    	
    SynWTFilter[][] filters=new SynWTFilter[2][];
    
    String [] words ;    

    words = word.split(",");
    
    if (words.length > 2)
    {
    	throw new IllegalArgumentException("Could not parse filter from " + word + ". [Too many commas]");
    }
    if (words.length == 1)
    {
    	filters[0] = parseHVFilters(words[0]);
    	filters[1] = filters[0];
    }
    else
    {
    	filters[0] = parseHVFilters(words[0]);
    	filters[1] = parseHVFilters(words[1]);
    }
    return filters;	 				

    }
    
    /**
     * Parse filters for one direction and multiple decomposition levels, seperated by ":" 
     * @param word
     * @return
     */
    private SynWTFilter[] parseHVFilters(String words)
    {
    	
    	String word;   	
    	StringTokenizer st = new StringTokenizer(words, ":");
    	SynWTFilter[] filt = new SynWTFilter[st.countTokens()];
    	
    	int i = 0;
    	while (st.hasMoreTokens())
    	{
    		word = st.nextToken();
    		filt[i] = parseFilter(word);
    		i++;    		
    	}
    	return filt;
    	
    }
    
    /**
     * parse a single filter from String word
     * @param word Filter specification
     * @return SynWTFilter corresponding to word
     */
    private SynWTFilter parseFilter (String word)
    {
    	SynWTFilter filt;
    	
    	if(word.equalsIgnoreCase("w5x3")){
    	    filt=new SynWTFilterIntLift5x3();
    	    return filt;
    	}
    	else if (word.length() > 6 && word.substring(0, 6).equalsIgnoreCase("whaarp")) {
    		float alpha;
    		try
    		{
    			alpha = Float.parseFloat(word.substring(6));
    		}
    		catch (Exception e)
    		{			
    			throw new IllegalArgumentException ("Could not parse alpha from \"" + word + "\"");
    		}   
    		if (plist.getBooleanParameter("symbolic")) {
    			filt=new SynWTParamFilterFloatHaarSym(alpha);
    		}
    		else {
    			filt=new SynWTParamFilterFloatHaar(alpha);
    		}
    		
    		return filt;
    	}
    	else if(word.equalsIgnoreCase("whaar")){
    		if (plist.getBooleanParameter("symbolic")) {
    			filt=new SynWTFilterFloatHaarSym();
    		}
    		else {
    			filt=new SynWTFilterFloatHaar();
    		}
    		return filt;
    	}
    	else if(word.equalsIgnoreCase("w9x7")){
    	    filt=new SynWTFilterFloatLift9x7();
    	    return filt;	 
    	}
    	else if (word.substring(0, 6).equalsIgnoreCase("w9x7pr")) {  
    		if (randomGen == null)
    		{
    			throw new IllegalArgumentException("w9x7pr is used, but no seed (key) for the random generator given");
    		}
    		filt = new SynWTParamFilterFloatLiftRandom9x7Factory (randomGen);
    		return filt;
    	}		
    	else if (word.substring(0, 5).equalsIgnoreCase("w9x7p")){				
    		float alpha;		
    		try
    		{
    			alpha = Float.parseFloat(word.substring(5));
    		}
    		catch (Exception e)
    		{			
    			throw new IllegalArgumentException ("Could not parse alpha from " + word);
    		}    	    	
    		if (plist.getBooleanParameter("symbolic"))
    		{
    			filt=new SynWTParamFilterFloatLift9x7Sym(alpha);
    		}
    		else
    		{
    			filt=new SynWTParamFilterFloatLift9x7(alpha);
    		}
    	    return filt;
    	}
    	else{
    	    throw new
    		IllegalArgumentException("Filter not defined: "
    					 +word);
    	}
    }


    
    
}
